from collections import Counter

import matplotlib
import pandas as pd
from flatbuffers.builder import np
from sklearn.datasets import fetch_20newsgroups

from chester.data_loader.webtext_data import load_data_pirates, load_data_king_arthur, load_data_chat_logs
from chester.run.full_run import run_madcat
from chester.run.user_classes import Data, ModelRun

matplotlib.use('TkAgg')
target_column = 'target'

################################################################################################
df1 = load_data_pirates().assign(target='pirate').sample(200, replace=True)
df2 = load_data_king_arthur().assign(target='arthur').sample(200, replace=True)
df3 = load_data_chat_logs().assign(target='chat').sample(100, replace=True)
df = pd.concat([df1, df2
                   , df3
                ])
# df['target'] = df['target'].apply(lambda x: 0 if "pirate" in x else 1)  # can do with or without
################################################################################################


################################################################################################
# url = "https://archive.ics.uci.edu/ml/machine-learning-databases/iris/iris.data"
# names = ['sepal-length', 'sepal-width', 'petal-length', 'petal-width', 'class']
# dataset = pd.read_csv(url, names=names)
# dataset.rename(columns={'class': 'target'}, inplace=True)
# df = dataset.sample(frac=1).reset_index(drop=True)
# df['target'] = df['target'].apply(lambda x: 0 if "Iris-setos" in x else 1)  # can do with or without
###############################################################################################

###############################################################################################
# Load the Boston Housing dataset. categorical
# boston = fetch_openml(name='boston', version=1)
# df = pd.DataFrame(boston.data, columns=boston.feature_names)
# df['target'] = boston.target
###############################################################################################

###############################################################################################
# df = pd.read_csv("chester/model_training/models/chester_models/day.csv")
# df.rename(columns={'cnt': 'target'}, inplace=True)
###############################################################################################

################################################################################################
from sklearn import datasets

digits = datasets.load_digits()
X = digits.images.reshape((len(digits.images), -1))
df = pd.DataFrame(X)
df.rename(columns={col: "feat_" + str(col) for col in df.columns}, inplace=True)
df['target'] = digits.target
df['target'] = "c_ " + df['target'].astype(str)


################################################################################################


################################################################################################
# categorical features
# import seaborn as sns
# df = sns.load_dataset("tips")
# df.rename(columns={'tip': target_column}, inplace=True)
################################################################################################

################################################################################################
# categorical
# import seaborn as sns
# df = sns.load_dataset("titanic")
# df.rename(columns={'survived': target_column}, inplace=True)
# df.drop(columns=['alive'], inplace=True)
###############################################################################################

###############################################################################################

def generate_data(n_features, n_rows, target_type='binary'):
    if target_type == 'binary':
        # Create binary target column
        target = np.random.choice(['yes', 'no'], size=n_rows)
    elif target_type == 'multiclass':
        # Create multiclass target column
        target = np.random.choice(['class_1', 'class_2', 'class_3', 'class_4',
                                   'class_5', 'class_6', 'class_7', 'class_8',
                                   'class_9', 'class_10', 'class_11', 'class_12'], size=n_rows)
    else:
        raise ValueError("Invalid target_type. Must be either 'binary' or 'multiclass'.")

    # Create feature categorical columns
    features = {}
    for i in range(n_features):
        feature = np.random.choice(['A', 'B', 'C', 'D'], size=n_rows)
        features[f'feature_{i}'] = feature

    # Create pandas DataFrame
    df = pd.DataFrame(features)
    df['target'] = target

    return df


# df = generate_data(20, 1000, target_type='binary')
# df = generate_data(30, 100, target_type='multiclass')
###############################################################################################

## vlad
def load_vlad():
    df = pd.read_csv("chester/model_training/models/chester_models/data.csv")
    df.rename(columns={'TOTAL_BET_AMOUNT': 'target'}, inplace=True)
    df['target'] = 1 * (df['REVENUE'] > 0.00001) + 1 * (df['REVENUE'] > 2)
    df.drop(
        columns=['REVENUE', 'Unnamed: 0', 'PLAYER_ID', 'MEDIAN_BET', 'SESSION_MINS', 'SPINS_COMPLETED', 'SPINS_STARTED',
                 'TOTAL_SPIN_LENGTH'], inplace=True)

    ## sample
    class_0 = df[df['target'] == 0].sample(5000)
    class_1 = df[df['target'] == 1]
    df = pd.concat([class_0, class_1])
    return df


def load_ex1():
    data = pd.read_csv("chester/model_training/models/chester_models/lead_df_2023-01-23.csv")
    data.rename(columns={'lawer_conversion_label': target_column}, inplace=True)
    data.drop(columns=["LEAD_ID"], inplace=True)
    return data


def load_ex2():
    newsgroups_train = fetch_20newsgroups(subset='train')
    df = pd.DataFrame(newsgroups_train.data, columns=['text'])
    y = newsgroups_train.target
    df['target'] = y
    category_counts = Counter(y)
    top_3_categories = category_counts.most_common(4)
    top_3_categories = [cat for cat, count in top_3_categories]
    df = df[df.target.isin(top_3_categories)].sample(1500)
    df['target'] = "c_ " + df['target'].astype(str)
    return df


import ml_datasets


def load_ex3():
    train_data, _ = ml_datasets.dbpedia()
    train_data = pd.DataFrame(train_data, columns=["text", "target"])
    train_data['target'] = "c_: " + train_data['target'].astype(str)
    return train_data


def load_ex4():
    train_data, _ = ml_datasets.cmu()
    train_data = pd.DataFrame(train_data, columns=["text", "target"])
    train_data['target'] = "category: " + train_data['target'].astype(str)
    train_data = train_data[train_data['target'].isin(train_data['target'].value_counts()[:5].index)]
    return train_data


def load_ex5():
    train_data, _ = ml_datasets.quora_questions()
    train_data = [(a, b, c) for ((a, b), c) in train_data]
    train_data = pd.DataFrame(train_data, columns=["text1", "text2", "target"])
    return train_data


# load data
# df = load_vlad()
# df = load_ex1()
# df = load_ex2()
# df = load_ex3().sample(1000)
# df = load_ex4().sample(1000)
# df = load_ex5().sample(900)


madcat_collector = run_madcat(Data(df=df, target_column='target'),
                              is_feature_stats=True,
                              is_pre_model=True,
                              model_run=ModelRun(n_models=3),
                              is_post_model=True, is_model_weaknesses=True
                              )



import matplotlib.pyplot as plt
import nltk
import pandas as pd
import seaborn as sns


def calculate_text_metrics(text):
    metrics = {'text_length': len(text), 'num_sentences': len(nltk.sent_tokenize(text))}
    if not text:
        return metrics
    if any(c.isalpha() for c in text):
        metrics['num_characters'] = len([c for c in text if c.isalpha()])
    else:
        metrics['num_characters'] = 0
    words = pd.Series(text.split())
    if len(words) > 0:
        metrics['num_words'] = len(words)
        metrics['num_unique_words'] = len(words.unique())
        metrics['ttr'] = metrics['num_unique_words'] / metrics['num_words']
    else:
        metrics['num_words'] = 0
        metrics['num_unique_words'] = 0
        metrics['ttr'] = 0
    metrics['num_missing'] = int(text == "")
    return metrics


def calculate_text_column_metrics(df, text_column='text'):
    df = df.copy()
    df[f'{text_column}_metrics'] = df[text_column].apply(calculate_text_metrics)
    words = df[text_column].str.split(expand=True).stack().unique()
    num_unique_words = len(words)
    df = pd.json_normalize(df[f'{text_column}_metrics'])
    return df, num_unique_words


def create_report(df, num_unique_words):
    report = ""
    report += f"Number of rows with missing data: {df['num_missing'].sum()}\n"
    report += f"Number of unique words: {num_unique_words}\n"
    report += f"Average type-token ratio: {df['ttr'].mean():.2f}\n"
    report += f"Median type-token ratio: {df['ttr'].median():.2f}\n"
    report += f"Average number of words per text: {df['num_words'].mean():.2f}\n"
    report += f"Median number of words per text: {df['num_words'].median():.2f}\n"
    report += f"Average number of sentences per text: {df['num_sentences'].mean():.2f}\n"
    report += f"Median number of sentences per text: {df['num_sentences'].median():.2f}\n"
    report += f"Average length of text: {df['text_length'].mean():.2f}\n"
    report += f"Median length of text: {df['text_length'].median():.2f}\n"
    return report


def plot_text_length_and_num_words(df):
    sns.set(style="darkgrid")
    fig, ax = plt.subplots(1, 2, figsize=(15, 10))
    sns.histplot(data=df, x='text_length', ax=ax[0])
    sns.histplot(data=df, x='num_words', ax=ax[1])
    fig.suptitle('Text Statistics')
    plt.show()


def analyze_text_stats(df, text_column='text'):
    df, num_unique_words = calculate_text_column_metrics(df, text_column)
    plot_text_length_and_num_words(df)
    report = create_report(df, num_unique_words)
    df_report = pd.DataFrame(
        data=[[row.split(":")[0], float(row.split(":")[1].strip())] for row in report.strip().split("\n")],
        columns=["Metric", "Value"])
    df_report["Value"] = df_report["Value"].apply(lambda x: f"{int(x)}" if x.is_integer() else f"{x:.2f}")
    return df_report


class TextAnalyzer:
    def __init__(self, df: pd.DataFrame = None, text_column: str = 'text', create_wordcloud: bool = True,
                 corex_topics: bool = True, key_sentences: bool = True,
                 common_words: bool = True, sentiment: bool = True,
                 ner_extraction: bool = True, kewords_extraction: bool = True,
                 data_quality: bool = True, corex_topics_num: int = 10,
                 top_words: int = 10, n_sentences: int = 5):
        self.df = df
        self.create_wordcloud = create_wordcloud
        self.corex_topics = corex_topics
        self.key_sentences = key_sentences
        self.common_words = common_words
        self.sentiment = sentiment
        self.data_quality = data_quality
        self.ner_extraction = ner_extraction
        self.kewords_extraction = kewords_extraction
        self.corex_topics_num = corex_topics_num
        self.top_words = top_words
        self.n_sentences = n_sentences
        self.text_column = text_column



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from wordcloud import WordCloud

from chester.model_analyzer.model_analysis import AnalyzeMessages
from chester.model_training.data_preparation import CVData
from chester.post_model_analysis.post_regression import VisualizeRegressionResults
from chester.zero_break.problem_specification import DataInfo


class PostModelAnalysis:
    def __init__(self, cv_data: CVData, data_info: DataInfo, model):
        self.cv_data = cv_data
        self.data_info = data_info
        self.model = model
        self.X_train = self.cv_data.train_data.drop(columns=[self.cv_data.target_column])
        self.y_train = self.cv_data.train_data[self.cv_data.target_column]
        self.X_test = self.cv_data.test_data.drop(columns=[self.cv_data.target_column])
        self.y_test = self.cv_data.test_data[self.cv_data.target_column]
        # retrain the model
        self.model.retrain(self.X_train, self.y_train)
        self.predict_test = self.model.predict(self.X_test)

    def plot_feature_importance(self, X_train: pd.DataFrame, top_feat=30):
        feature_importance = self.model.model.feature_importances_
        feature_importance = 100.0 * (feature_importance / feature_importance.max())
        feature_names = X_train.columns
        important_idx = np.argsort(feature_importance)
        data = {'feature_names': feature_names[important_idx],
                'feature_importance': feature_importance[important_idx]}
        df = pd.DataFrame(data)
        print(AnalyzeMessages().feature_importance_message())
        df.sort_values(by='feature_importance', ascending=False, inplace=True)
        sns.barplot(x='feature_importance', y='feature_names', data=df[0:top_feat])
        plt.figure(figsize=(15, 13))
        plt.title(f'Feature Importance (top {top_feat} features)')
        self.plot_wordcloud_importance(df=df)
        plt.show()

    @staticmethod
    def plot_wordcloud_importance(df, title="Feature Importance Cloud"):
        """
        Plot word cloud of features weighted by their importance.
        :param features_importance: List of tuples (feature name, feature importance).
        :param title: Title of the plot.
        :return: None.
        """
        df = df.sort_values(by='feature_importance', ascending=False)
        features_importance = list(zip(df['feature_names'], df['feature_importance']))

        wordcloud = WordCloud(
            random_state=21,
            normalize_plurals=True,
            max_words=100).generate_from_frequencies(dict(features_importance))
        plt.figure(figsize=(10, 10))
        plt.imshow(wordcloud)
        plt.axis("off")
        plt.title(title, fontsize=15)
        plt.show()

    def shap_values(self, X_train: pd.DataFrame, shap):
        explainer = shap.Explainer(self.model, X_train, check_additivity=False)
        shap_values = explainer(X_train, check_additivity=False)
        plt.title("SHAP values for train set")
        print(AnalyzeMessages().shap_values_message())
        shap.summary_plot(shap_values, X_train)

    def analyze(self,
                shap_values: bool = True,
                coefficients: bool = True,
                confusion_matrix: bool = True,
                roc_curve: bool = True,
                learning_curve: bool = True,
                feature_importance: bool = True,
                regression_visual: bool = True) -> None:
        if 'regression' in self.data_info.problem_type_val.lower():
            if regression_visual:
                VisualizeRegressionResults(self.y_test, self.predict_test).all_plots()
        if feature_importance:
            try:
                self.plot_feature_importance(self.X_train)
            except:
                pass
        if shap_values:
            try:
                import shap as shp
                self.shap_values(self.X_train, shp)
            except:
                pass
        if coefficients:
            try:
                self.coefficients()
            except:
                pass
        if confusion_matrix:
            if 'class' in self.data_info.problem_type_val.lower():
                try:
                    self.confusion_matrix(self.X_test, self.y_test)
                except:
                    pass
        if learning_curve:
            self.learning_curve(self.X_train, self.y_train)
        if roc_curve:
            try:
                self.roc_curve(self.X_test, self.y_test)
            except:
                try:
                    self.roc_curve_multiclass(self.X_test, self.y_test)
                except:
                    pass

    def confusion_matrix(self, X_test: pd.DataFrame, y_test: pd.Series) -> None:
        from sklearn.metrics import confusion_matrix
        import seaborn as sns
        print(AnalyzeMessages().confusion_matrix_message())

        y_pred = self.model.predict(X_test)
        cm = confusion_matrix(y_test, y_pred)
        sns.heatmap(cm, annot=True, cmap='Blues', fmt='g')
        plt.title('Confusion Matrix')
        plt.xlabel('Predicted')
        plt.ylabel('Actual')
        plt.show()

    def roc_curve(self, X_test: pd.DataFrame, y_test: pd.Series) -> None:
        from sklearn.metrics import roc_curve, auc
        try:
            y_pred = self.model.predict(X_test)
            fpr, tpr, thresholds = roc_curve(y_test, y_pred)
            roc_auc = auc(fpr, tpr)
            plt.figure()
            plt.plot(fpr, tpr, color='darkorange', lw=1, label='ROC curve (area = %0.2f)' % roc_auc)
            plt.plot([0, 1], [0, 1], color='navy', lw=1, linestyle='--')
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('Receiver operating characteristic')
            plt.legend(loc="lower right")
            print(AnalyzeMessages().roc_curve_message())
            plt.show()
        except:
            pass

    def roc_curve_multiclass(self, X_test: pd.DataFrame, y_test: pd.Series) -> None:
        from sklearn.metrics import roc_auc_score, roc_curve
        from sklearn.preprocessing import label_binarize
        from scipy import interp

        y_test_bin = label_binarize(y_test, classes=np.unique(y_test))
        y_pred = self.model.predict_proba(X_test)

        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(y_test_bin.shape[1]):
            fpr[i], tpr[i], _ = roc_curve(y_test_bin[:, i], y_pred[:, i])
            roc_auc[i] = roc_auc_score(y_test_bin[:, i], y_pred[:, i])

        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(y_test_bin.shape[1])]))

        mean_tpr = np.zeros_like(all_fpr)
        for i in range(y_test_bin.shape[1]):
            mean_tpr += interp(all_fpr, fpr[i], tpr[i])

        mean_tpr /= y_test_bin.shape[1]
        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = roc_auc_score(y_test_bin, y_pred, average='macro')

        plt.figure()
        plt.plot(fpr["macro"], tpr["macro"], color='darkorange', lw=1, label='Macro-average ROC curve (area = {0:0.2f})'
                                                                             ''.format(roc_auc["macro"]))

        for i in range(y_test_bin.shape[1]):
            plt.plot(fpr[i], tpr[i], lw=1, label='ROC curve of class {0} (area = {1:0.2f})'
                                                 ''.format(i, roc_auc[i]))

        plt.plot([0, 1], [0, 1], color='navy', lw=1, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver operating characteristic')
        plt.legend(loc="lower right")
        print(AnalyzeMessages().roc_curve_message())
        plt.show()

    def learning_curve(self, X: pd.DataFrame, y: pd.Series) -> None:
        import os
        import sys

        original_stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')
        try:
            from sklearn.model_selection import learning_curve
            train_sizes, train_scores, test_scores = learning_curve(self.model.model, X, y, cv=5, scoring='accuracy')
            train_scores_mean = np.mean(train_scores, axis=1)
            train_scores_std = np.std(train_scores, axis=1)
            test_scores_mean = np.mean(test_scores, axis=1)
            test_scores_std = np.std(test_scores, axis=1)
            plt.plot(train_sizes, train_scores_mean, 'o-', color="r", label="Training score")
            plt.plot(train_sizes, test_scores_mean, 'o-', color="g", label="Cross-validation score")
            plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                             train_scores_mean + train_scores_std, alpha=0.1, color="r")
            plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                             test_scores_mean + test_scores_std, alpha=0.1, color="g")
            plt.grid()
            plt.xlabel("Training examples")
            plt.ylabel("Accuracy Score")
            plt.title("Learning Curve")
            plt.legend(loc="best")
            print(AnalyzeMessages().learning_curve_message())
            plt.show()
        except:
            pass
        finally:
            sys.stderr = original_stderr

    def coefficients(self) -> None:
        try:
            coef = self.model.model.coef_[0]
            print(coef)
            if len(coef) < 10:
                plt.bar(np.arange(len(coef)), coef)
                plt.title("Coefficients for logistic regression model")
                plt.xlabel("Features")
                plt.ylabel("Values")
            else:
                sns.violinplot(coef, inner="stick")
                plt.title("Coefficients distribution for logistic regression model")
                plt.xlabel("Coefficients")
                print(AnalyzeMessages().coefficients_message())
            plt.show()
        except:
            pass



import re
import unicodedata

import nltk
from nltk.corpus import stopwords


def get_stopwords():
    nltk.download('stopwords', quiet=True)
    return stopwords.words('english')


def remove_html(text: str) -> str:
    """Remove HTML tags from a string.

    Args:
        text: The input string.

    Returns:
        The input string with HTML tags removed.
    """
    return re.sub(r'<[^>]*>', '', text)


def remove_accents(text: str) -> str:
    """Remove accents from characters in a string.

    Args:
        text: The input string.

    Returns:
        The input string with accents removed.
    """
    text = ''.join(c for c in unicodedata.normalize('NFD', text) if unicodedata.category(c) != 'Mn')
    return text


def expand_contractions(text: str, contraction_mapping: dict) -> str:
    """Expand contractions in a string.

    Args:
        text: The input string.
        contraction_mapping: A dictionary mapping contractions to their expanded form.

    Returns:
        The input string with contractions expanded.
    """
    contractions_pattern = re.compile('({})'.format('|'.join(contraction_mapping.keys())),
                                      flags=re.IGNORECASE | re.DOTALL)

    def expand_match(contraction):
        match = contraction.group(0)
        first_char = match[0]
        expanded_contraction = contraction_mapping.get(match) if contraction_mapping.get(
            match) else contraction_mapping.get(match.lower())
        expanded_contraction = first_char + expanded_contraction[1:]
        return expanded_contraction

    expanded_text = contractions_pattern.sub(expand_match, text)
    expanded_text = re.sub("'", "", expanded_text)
    return expanded_text


contractions_dict = {
    "ain't": "am not / are not / is not / has not / have not",
    "aren't": "are not / am not",
    "can't": "cannot",
    "can't've": "cannot have",
    "'cause": "because",
    "could've": "could have",
    "couldn't": "could not",
    "couldn't've": "could not have",
    "didn't": "did not",
    "doesn't": "does not",
    "don't": "do not",
    "hadn't": "had not",
    "hadn't've": "had not have",
    "hasn't": "has not",
    "haven't": "have not",
    "he'd": "he had / he would",
    "he'd've": "he would have",
    "he'll": "he shall / he will",
    "he'll've": "he shall have / he will have",
    "he's": "he has / he is",
    "how'd": "how did",
    "how'd'y": "how do you",
    "how'll": "how will",
    "how's": "how has / how is / how does",
    "I'd": "I had / I would",
    "I'd've": "I would have",
    "I'll": "I shall / I will",
    "I'll've": "I shall have / I will have",
    "I'm": "I am",
    "I've": "I have",
    "isn't": "is not",
    "it'd": "it had / it would",
    "it'd've": "it would have",
    "it'll": "it shall / it will",
    "it'll've": "it shall have / it will have",
    "it's": "it has / it is",
    "let's": "let us",
    "ma'am": "madam",
    "mayn't": "may not",
    "might've": "might have",
    "mightn't": "might not",
    "mightn't've": "might not have",
    "must've": "must have",
    "mustn't": "must not",
    "mustn't've": "must not have",
    "needn't": "need not",
    "needn't've": "need not have",
    "o'clock": "of the clock"}


def remove_non_ascii(text):
    """Remove non-ASCII characters from the given text."""
    text = ''.join([c for c in text if 0 < ord(c) < 127])
    return text


def remove_numbers(text):
    """Remove numbers from the given text."""
    text = ''.join([c for c in text if not c.isdigit()])
    return text



from chester.cleaning import cleaning_func as cln

from chester.zero_break.problem_specification import DataInfo


class CleanerHandler:
    def __init__(self, data_info: DataInfo, text_cleaner: cln.TextCleaner = None):
        self.data_info = data_info
        self.text_cleaner = text_cleaner
        if text_cleaner is not None:
            self.text_cleaner.df = self.data_info.data
        self.target = data_info.target
        self.problem_type_val = data_info.problem_type_val
        self.feature_types_val = data_info.feature_types_val
        self.loss_detector_val = data_info.loss_detector_val
        self.metrics_detector_val = data_info.metrics_detector_val
        self.model_selection_val = data_info.model_selection_val
        self.label_transformation_val = data_info.label_transformation_val

    def transform(self):
        text_columns = self.feature_types_val.get("text")
        for col in text_columns:
            if self.text_cleaner is not None:
                curr_text_cleaner = self.text_cleaner
                curr_text_cleaner.text_column = col
                text_cleaner = curr_text_cleaner
            else:
                text_cleaner = cln.TextCleaner(self.data_info.data, text_column=col)
            print(f"{col} text cleaning")
            text_cleaner.generate_report()
            self.data_info.data = cln.clean_text_df(text_cleaner)



import pandas as pd

from chester.text_stats_analysis import common_words as cw, corex_topics as corex, keywords_extraction as kw_extract
from chester.text_stats_analysis.data_quality import analyze_text_stats, TextAnalyzer
from chester.text_stats_analysis.key_sentences import extract_key_sentences
from chester.text_stats_analysis.sentiment import analyze_sentiment, report_sentiment_stats, plot_sentiment_scores
from chester.text_stats_analysis.word_cloud import create_word_cloud

data_quality_message = "Before analyzing text, ensure data is clean and of good quality.\n" \
                       "Report provides key stats on data quality, incl. missing data,\n" \
                       "unique words, type-token ratio, words/sentences per text, and text length:"

common_words_message = "Count word frequency for basic text analysis.\n" \
                       "Report shows most common words, giving insight into text topic and content:"

word_cloud_message = "A word cloud visualizes most common words in text. Size indicates frequency.\n" \
                     "It quickly identifies text data's main themes and topics.\n"

sentiment_analysis_message = "\nSentiment analysis determines emotional tone of text.\n" \
                             "Report shows sentiment breakdown: positive, negative, neutral:"

corex_topic_message = "Corex topic modeling identifies latent topics in text data.\n" \
                      "Report shows top topics, each topic represented by the following words:"

key_sentences_message = "Extracting key sentences identifies important/representative sentences in text.\n" \
                        "Report shows top sentences for each text topic:\n"

kws_message = "Extracting key words identifies important/representative words in text.\n" \
              "Report shows top words for each text topic:"


def print_analyze_message(create_wordcloud: bool = True,
                          corex_topics: bool = True,
                          key_sentences: bool = True,
                          common_words: bool = True,
                          sentiment: bool = True,
                          data_quality: bool = True,
                          corex_topics_num: int = 10,
                          top_words: int = 10,
                          n_sentences: int = 5):
    order_of_operations = ["First,", "Next,", "Then,", "Additionally,", "Furthermore,", "Later,"]
    operations_counter = 0
    if data_quality:
        print(
            f"{order_of_operations[operations_counter]} "
            f"we will analyze text statistics such as the number of words, unique words and type-token ratio.")
        operations_counter += 1
    if create_wordcloud:
        print(
            f"{order_of_operations[operations_counter]} "
            f"we will create a wordcloud to visualize the most common words in the text.")
        operations_counter += 1
    if corex_topics:
        print(
            f"{order_of_operations[operations_counter]} "
            f"we will extract {corex_topics_num} key topics using Corex and present them with their top words.")
        operations_counter += 1
    if key_sentences:
        print(f"{order_of_operations[operations_counter]} "
              f"we will identify key sentences in the text.")
        operations_counter += 1
    if common_words:
        print(
            f"{order_of_operations[operations_counter]} "
            f"we will extract the {top_words} most common words from the text.")
        operations_counter += 1
    if sentiment:
        print(f"{order_of_operations[operations_counter]} "
              f"we will analyze the sentiment of the text.")
        operations_counter += 1
    print("The results of text cleaning and preprocessing, including key words and key sentences,\n"
          "may not accurately reflect the original content and could lead to incorrect conclusions,\n "
          "if not handled properly.\n")


def analyze_text(df: pd.DataFrame,
                 create_wordcloud: bool = True,
                 corex_topics: bool = True,
                 key_sentences: bool = True,
                 common_words: bool = True,
                 sentiment: bool = True,
                 data_quality: bool = True,
                 kewords_extraction: bool = True,
                 corex_topics_num: int = 10,
                 top_words: int = 10,
                 n_sentences: int = 10,
                 text_column: str = 'text'):
    """
    Analyze text using various text analysis techniques.
    :param df: pandas data with clean text column
    :param create_wordcloud: flag indicating whether to create a word cloud (default: True)
    :param corex_topics: flag indicating whether to extract corex topics (default: True)
    :param key_sentences: flag indicating whether to extract key sentences (default: True)
    :param common_words: flag indicating whether to extract common words (default: True)
    :param sentiment: flag indicating whether to perform sentiment analysis (default: True)
    :param data_quality: flag indicating whether to check data quality (default: True)
    :param kewords_extraction: flag indicating whether to create potential keywords
    :param corex_topics_num: number of corex topics to extract (default: 10)
    :param top_words: top words
    :param n_sentences: number of sentences to return
    """

    is_clean_col_exists = True if f"clean_{text_column}" in df.columns else False
    if is_clean_col_exists:
        modified_df = df.drop(columns=[text_column], axis=1). \
            rename(columns={f'clean_{text_column}': text_column})

    print_analyze_message()
    if data_quality:
        print(data_quality_message)
        if is_clean_col_exists:
            print(analyze_text_stats(modified_df, text_column), "\n")
        else:
            print(analyze_text_stats(df, text_column), "\n")
    if common_words:
        print(common_words_message)

        if is_clean_col_exists:
            print("\t", cw.most_common_words(modified_df, common_words=top_words, text_column=text_column))
        else:
            print("\t", cw.most_common_words(df, common_words=top_words, text_column=text_column))

        print("\n")
    if create_wordcloud:
        print(word_cloud_message)

        if is_clean_col_exists:
            create_word_cloud(modified_df, text_column=text_column)
        else:
            create_word_cloud(df, text_column=text_column)

    if sentiment:
        print(sentiment_analysis_message)
        df = df.copy()
        modified_df = modified_df.copy()
        if is_clean_col_exists:
            df = analyze_sentiment(modified_df, text_column=text_column)
        else:
            df = analyze_sentiment(df, text_column=text_column)
        print(report_sentiment_stats(df), "\n")
        print()
        plot_sentiment_scores(df)

    if corex_topics:
        print(corex_topic_message)

        if is_clean_col_exists:
            corex.plot_corex_wordcloud(modified_df, n_topics=corex_topics_num, top_words=top_words,
                                       text_column=text_column)
        else:
            corex.plot_corex_wordcloud(df, n_topics=corex_topics_num, top_words=top_words, text_column=text_column)

    if key_sentences:
        print(key_sentences_message)

        if is_clean_col_exists:
            print(extract_key_sentences(modified_df, n_sentences=n_sentences, text_column=text_column))
        else:
            print(extract_key_sentences(df, n_sentences=n_sentences, text_column=text_column))

    if kewords_extraction:
        print(kws_message)

        if is_clean_col_exists:
            full_text = '. '.join(modified_df[text_column])
        else:
            full_text = '. '.join(df[text_column])

        kws_rake = kw_extract.RakeKeywordExtractor()
        print("\t", kws_rake.extract(text=full_text))


def analyze_text_df(text_analyzer: TextAnalyzer):
    df = text_analyzer.df
    analyze_text(df=df, text_column=text_analyzer.text_column,
                 create_wordcloud=text_analyzer.create_wordcloud,
                 corex_topics=text_analyzer.corex_topics,
                 key_sentences=text_analyzer.key_sentences,
                 common_words=text_analyzer.common_words,
                 sentiment=text_analyzer.sentiment,
                 data_quality=text_analyzer.data_quality,
                 corex_topics_num=text_analyzer.corex_topics_num,
                 top_words=text_analyzer.top_words,
                 n_sentences=text_analyzer.n_sentences)



import math
import random
from math import floor

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch
from sklearn.manifold import TSNE
from wordcloud import WordCloud

from chester.zero_break.problem_specification import DataInfo


class CategoricalPreModelAnalysis:
    def __init__(self, data_info: DataInfo):
        self.data_info = data_info
        self.cols = self.data_info.feature_types_val["categorical"]
        self.n_cols = len(self.cols)
        self.target = self.data_info.data[self.data_info.target]
        self.target_df = self.data_info.data[[self.data_info.target]]
        self.data = self.data_info.data[self.cols]
        # calc
        self.cols_sorted = self.sort_by_pvalue()
        self.cols_sorted_with_pvalue = None

    def tsne(self):
        X_tsne_3d = TSNE(n_components=3).fit_transform(pd.get_dummies(self.data))
        X_tsne_2d = X_tsne_3d[:, :2]

        fig = plt.figure(figsize=(16, 10))
        ax1 = plt
        # ax2 = fig.add_subplot(122, projection='3d')
        if self.data_info.problem_type_val in ["Regression"]:
            ax1.hexbin(X_tsne_2d[:, 0], X_tsne_2d[:, 1], C=self.target, gridsize=50, cmap='viridis', edgecolors='black')
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=self.target, cmap='viridis')
            ax1.title("Visualizing Categorical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Categorical Features and Target with t-SNE (3D)")
        elif self.data_info.problem_type_val in ["Binary regression", "Binary classification"]:
            target_classes = self.target.unique()
            color_map = {target_class: color for target_class, color in zip(target_classes, ['red', 'blue'])}
            colors = self.target.apply(lambda x: color_map[x])
            ax1.scatter(X_tsne_2d[:, 0], X_tsne_2d[:, 1], c=colors)
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=colors)
            legend_handles = [Patch(color=color_map[target_class], label=target_class) for target_class in
                              target_classes]
            ax1.title("Visualizing Categorical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Categorical Features and Target with t-SNE (3D)")
            ax1.legend(handles=legend_handles)
        else:  # Multi-class classification
            target_classes = self.target.unique()
            color_map = {target_class: color for target_class, color in
                         zip(target_classes, plt.cm.rainbow(np.linspace(0, 1, len(target_classes))))}
            ax1.legend(
                handles=[Patch(color=color_map[target_class], label=target_class) for target_class in target_classes])
            # ax2.legend(
            #     handles=[Patch(color=color_map[target_class], label=target_class) for target_class in target_classes])

    def any_categorical(self):
        return True if len(self.cols) > 0 else False

    def sort_by_pvalue(self):
        from sklearn.cluster import KMeans
        from scipy.stats import chi2_contingency
        import warnings
        warnings.simplefilter("ignore")

        problem_type = self.data_info.problem_type_val

        if problem_type in ["Regression"]:
            num_groups = min(floor(self.data_info.rows / 20), 10)
            kmeans = KMeans(n_clusters=num_groups, n_init=10)
            kmeans.fit(self.target_df)
            target_labels = kmeans.labels_

        feature_pvalue_list = []
        for col in self.cols:
            data_col = self.data[col]
            if problem_type == "Regression":
                contingency_table = pd.crosstab(data_col, columns=target_labels)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            else:
                contingency_table = pd.crosstab(index=data_col, columns=self.target)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            feature_pvalue_list.append((col, pvalue))

        sorted_list = sorted(feature_pvalue_list, key=lambda x: x[1], reverse=False)
        self.cols_sorted_with_pvalue = sorted_list
        return [x[0] for x in sorted_list]

    def analyze_pvalue(self, is_plot=True, top_features=10):
        self.sort_by_pvalue()
        if len(self.cols) == 0:
            return None
        self.plot_wordcloud_pvalues(self.cols_sorted_with_pvalue)
        if is_plot:
            if self.n_cols > 50:
                print("plotting!")
                self.plot_histogram_pvalues(self.cols_sorted_with_pvalue)
        print("Pvalues for top categorical features:")
        print(pd.DataFrame(self.cols_sorted_with_pvalue[0:top_features], columns=["feature", "pvalue"]))

    def partial_plot(self, classification_row_percent=True):
        import warnings
        warnings.simplefilter("ignore")
        top_features = 25

        if self.n_cols <= 25:
            sample_features = self.n_cols
            top_features = self.n_cols
        else:
            sample_features = min(2 * 25, int(self.n_cols / 2))
        top_feature_names = random.sample(self.cols_sorted[0:top_features], sample_features)
        feature_index = {feature: index for index, feature in enumerate(self.cols_sorted[0:top_features])}
        top_feature_names.sort(key=lambda x: feature_index[x])
        if self.data_info.problem_type_val in ["Binary regression", "Binary classification"]:
            max_plots = 9
            top_n = self.data[:top_features].columns
            dim = math.ceil(math.sqrt(len(top_n)))
            num_rows = math.ceil(max_plots / dim)
            fig, ax = plt.subplots(dim, dim)
            fig.tight_layout()
            if classification_row_percent:
                plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                             "Showing % from Feature (row)",
                             fontsize=14, fontweight='bold')
            else:
                plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                             "Showing % from Target (column)",
                             fontsize=14, fontweight='bold')
            plt.figure(figsize=(15, 15))
            for i, col in enumerate(top_feature_names):
                if i >= num_rows * num_rows:
                    return None
                ax_i = ax[i // dim, i % dim]
                if classification_row_percent:
                    crosstab = pd.crosstab(self.data[col], self.target, normalize='index') * 100
                    crosstab = crosstab[(crosstab.T != 0).any()]
                    crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                    crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                else:
                    crosstab = pd.crosstab(self.data[col], self.target, normalize='columns') * 100
                    crosstab = crosstab[(crosstab.T != 0).any()]
                    crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                    crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                sns.heatmap(crosstab, annot=False, cmap="YlGnBu", fmt='g', ax=ax_i)
                ax_i.set_ylabel(None)
                ax_i.set_xlabel(None)
                ax_i.set_title(col, fontsize=12, fontweight='bold')
            plt.tight_layout()
            plt.show()
        if self.data_info.problem_type_val in ["Regression"]:
            from sklearn.cluster import KMeans
            target = self.target
            kmeans = KMeans(n_clusters=10)
            kmeans.fit(self.target_df)
            target_labels = kmeans.labels_
            plt.figure(figsize=(12, 12))
            plt.suptitle("Partial Plot to Identify Patterns between Categorical Sampled Features and Target",
                         fontsize=16,
                         fontweight='bold')
            grid_size = 4
            num_features = min(grid_size * grid_size, top_features)
            num_rows = int(np.ceil(num_features / grid_size))
            for i, col in enumerate(top_feature_names[:num_features]):
                plt.subplot(num_rows, grid_size, i + 1)
                column = self.data[col]
                if column.dtype == "object":
                    column = column.astype("category").cat.codes
                    column = column[column < 5]
                crosstab = pd.crosstab(target_labels, column, normalize='index')
                sns.heatmap(crosstab, annot=False, cmap='Greens', cbar=True)
                plt.ylabel("Clusters")
                plt.xlabel("{} Value".format(col))
                plt.subplots_adjust(hspace=0.5, wspace=0.5)
        elif self.data_info.problem_type_val in ["Multiclass classification"]:
            plt.figure(figsize=(16, 16))
            plt.suptitle(
                "Heatmap to Show Correlation between Sampled Categorical Features (top 5 categories) and Target",
                fontsize=16,
                fontweight='bold')
            top_features = 16
            top_feature_names = self.data[:top_features].columns
            rows = 4
            cols = 4
            for i, col in enumerate(top_feature_names):
                if i > 15:
                    return None
                plt.subplot(rows, cols, i + 1)
                crosstab = pd.crosstab(self.data[col], self.target, normalize='index') * 100
                crosstab = crosstab[(crosstab.T != 0).any()]
                crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                sns.heatmap(crosstab, annot=False, cmap="YlGnBu", fmt='g')
                plt.title(col, fontsize=12, fontweight='bold')
            plt.show()
            return None

    @staticmethod
    def plot_histogram_pvalues(features_pvalues):
        """
        Plot histogram of p-values for features.
        :param features_pvalues: List of tuples (column name, pvalue).
        :return: None.
        """
        pvalues = [pvalue for _, pvalue in features_pvalues]
        fig, ax = plt.subplots()
        ax.hist(pvalues, bins=50, edgecolor='k', color='#2ecc71')
        ax.set_title("Histogram of P-values for Categorical Features", fontsize=16)
        ax.set_xlabel("P-value", fontsize=14)
        ax.set_ylabel("Frequency", fontsize=14)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_linewidth(0.5)
        ax.spines['bottom'].set_linewidth(0.5)
        ax.tick_params(axis='both', which='both', labelsize=12)
        plt.show()
        return None

    @staticmethod
    def plot_wordcloud_pvalues(features_pvalues,
                               title="Categorical Features Pvalues Based on Partial Plot"):
        """
        Plot word cloud of features weighted by their p-value.
        :param features_pvalues: List of tuples (column name, pvalue).
        :param title: Title of the plot.
        :return: None.
        """
        features_pvalues = [(feature, 1 - pvalue) for feature, pvalue in features_pvalues]
        wordcloud = WordCloud(
            random_state=21,
            normalize_plurals=True).generate_from_frequencies(dict(features_pvalues))
        plt.imshow(wordcloud)
        plt.axis("off")
        plt.title(title, fontsize=15)
        plt.show()
        return None

    def run(self, is_plot=True):
        if self.n_cols > 1:
            self.analyze_pvalue(is_plot=is_plot)
            if is_plot:
                if 'classification' or 'binary regression' in self.data_info.problem_type_val.lower():
                    self.partial_plot(classification_row_percent=False)
                    self.partial_plot(classification_row_percent=True)
                else:
                    self.partial_plot()
                self.tsne()
        elif self.n_cols == 1:
            self.analyze_pvalue(is_plot=is_plot)
            if is_plot:
                self.partial_plot()
        return None


def format_df(df, max_value_width=10, ci_max_value_width=15, ci_col="CI"):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    def trim_ci_value(val):
        if len(str(val)) > ci_max_value_width:
            return str(val)[:ci_max_value_width] + "..."
        return str(val)

    df_subset = df.drop(ci_col, axis=1)
    df_subset = df_subset.applymap(trim_value)
    df[df_subset.columns] = df_subset
    df[ci_col] = df[ci_col].apply(trim_ci_value)

    return df



import matplotlib.pyplot as plt
import scipy.stats as stats


class VisualizeRegressionResults:
    def __init__(self, y_true, y_pred):
        self.y_true = y_true
        self.y_pred = y_pred

    def scatter_plot(self, ax):
        ax.scatter(self.y_true, self.y_pred, color='blue', alpha=0.5)
        ax.set_xlabel('True Values')
        ax.set_ylabel('Predictions')
        ax.set_title('Scatter Plot of True Values vs Predictions')

    def residual_plot(self, ax):
        residuals = self.y_true - self.y_pred
        ax.scatter(self.y_pred, residuals, color='blue', alpha=0.5)
        ax.set_xlabel('Predictions')
        ax.set_ylabel('Residuals')
        ax.set_title('Residual Plot')
        ax.hlines(y=0, xmin=min(self.y_pred), xmax=max(self.y_pred), color='red', linestyle='--')

    def histogram_plot(self, ax):
        residuals = self.y_true - self.y_pred
        ax.hist(residuals, bins=50, color='blue', alpha=0.5)
        ax.set_xlabel('Residuals')
        ax.set_ylabel('Count')
        ax.set_title('Histogram of Residuals')

    def qq_plot(self, ax):
        residuals = self.y_true - self.y_pred
        stats.probplot(residuals, dist='norm', plot=ax)
        ax.set_title('Q-Q Plot of Residuals')

    def all_plots(self):
        fig, axs = plt.subplots(2, 2, figsize=(15, 15))
        self.scatter_plot(axs[0, 0])
        self.residual_plot(axs[0, 1])
        self.histogram_plot(axs[1, 0])
        self.qq_plot(axs[1, 1])
        plt.show()



import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from chester.model_analyzer import model_analysis

analysis_message = "The following graph displays boxplots of the calculated metrics for each fold\n, " \
                   "with the x-axis representing the fold number and the y-axis representing the metric value.\n " \
                   "The boxplot shows the median, interquartile range, and outliers of the metric values \n for both " \
                   "train and test data\n. " \
                   "A large difference between the train and test boxplots may indicate overfitting."


def analyze_results(results: pd.DataFrame, parameters):
    # remove parameters columns from the results dataframe
    results_copy = results.drop([p.name for p in parameters], axis=1)
    # check if there are any lists in the dataframe
    if any(isinstance(i, list) for i in results_copy.values):
        # iterate through the dataframe and flatten any lists
        results_copy = results_copy.applymap(lambda x: x[0] if isinstance(x, list) else x)
    # keep only numeric columns and id_vars=['type', 'fold']
    results_copy = pd.concat([
        results_copy[['type', 'fold']]
        , results_copy.select_dtypes(include=['float64', 'int64'])]
        , axis=1)
    # melt the dataframe to have (type, fold, metric, value) format
    results_copy = results_copy.loc[:, ~results_copy.columns.duplicated()]
    results_melt = pd.melt(results_copy, id_vars=['type', 'fold'], value_vars=results_copy.columns[2:])
    results_melt = results_melt.astype({'fold': 'int64'})
    results_agg = results_melt.groupby(['type', 'variable']).agg({'value': 'mean'}).reset_index()
    results_pivot = results_agg.pivot(index='type', columns='variable', values='value')
    pd.set_option('display.max_rows', None)
    pd.set_option('display.width', None)
    print("Model performance:")
    print(model_analysis.AnalyzeMessages().performance_metrics_message())
    print(results_pivot)

    sns.boxplot(x='fold', y='value', hue='variable', data=results_melt, palette="Set3")
    print(analysis_message)
    plt.show()



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from chester.model_monitor import calculate_scores_utils
from chester.model_training.data_preparation import CVData
from chester.model_training.models.chester_models.catboost.catboost_utils import calculate_catboost_metrics_scores
from chester.zero_break.problem_specification import DataInfo

boostrap_message = "The following plots display the distribution of the metrics, " \
                   "\nusing a violin plot and histogram. " \
                   "\n\tThe violin plot shows the median, " \
                   "quartiles, and other important summary statistics of the data, " \
                   "\n\twhile the histogram provides a visual representation of the frequency of values in the data. " \
                   "\nThese plots give you a sense of how the metrics are distributed\n " \
                   "and can be used to identify any " \
                   "patterns or outliers in the data."


class ModelBootstrap:
    def __init__(self, cv_data: CVData, data_info: DataInfo, model):
        self.cv_data = cv_data
        self.data_info = data_info
        self.model = model
        self.X_train = self.cv_data.train_data.drop(columns=[self.cv_data.target_column])
        self.y_train = self.cv_data.train_data[self.cv_data.target_column]
        self.X_test = self.cv_data.test_data.drop(columns=[self.cv_data.target_column])
        self.y_test = self.cv_data.test_data[self.cv_data.target_column]
        # retrain the model
        self.model.retrain(self.X_train, self.y_train)
        self.predict_test = self.model.predict(self.X_test)
        self.metrics = self.get_metrics_functions()

    def get_metrics_functions(self):
        metric_functions = []
        metrics = self.data_info.metrics_detector_val
        for metric in metrics:
            metric_function = self.get_metric_function(metric)
            if metric_function is not None:
                metric_functions.append(metric_function)
        return metric_functions

    def get_metric_function(self, metric):
        from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
        from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_absolute_percentage_error
        from sklearn.metrics import roc_auc_score

        if metric == 'Accuracy':
            return accuracy_score
        elif metric == 'Precision':
            return precision_score
        elif metric == 'Recall':
            return recall_score
        elif metric == 'F1':
            return f1_score
        elif metric == "MSE":
            return mean_squared_error
        elif metric == "MAE":
            return mean_absolute_error
        elif metric == "MAPE":
            return mean_absolute_percentage_error
        elif metric == 'ROC':
            return roc_auc_score
        else:
            return None

    def bootstrap_metrics(self, B=200, sample=200):
        n = len(self.y_test)
        bootstrap_metrics = []
        for i in range(B):
            sample_indexes = np.random.randint(0, n, sample)
            X_sample = self.X_test.iloc[sample_indexes]
            y_sample = self.y_test.iloc[sample_indexes]
            y_pred = self.model.predict(X_sample)

            try:
                bootstrap_metrics.append(
                    calculate_catboost_metrics_scores(
                        y_sample, prediction=y_pred, metrics_list=self.metrics,
                        problem_type=self.data_info.problem_type_val)
                )
            except:
                bootstrap_metrics.append(
                    calculate_scores_utils.calculate_metrics_scores(
                        y=y_sample, prediction=y_pred, metrics_list=self.metrics,
                        problem_type=self.data_info.problem_type_val)
                )
        return bootstrap_metrics

    def plot(self):
        print(boostrap_message)
        import seaborn as sns
        metrics = pd.DataFrame(self.bootstrap_metrics())
        for metric_name in metrics.columns:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 5))
            sns.violinplot(x=metrics[metric_name], ax=ax1)
            ax1.set_title(metric_name)
            sns.histplot(x=metrics[metric_name], ax=ax2)
            plt.show()



import math

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from chester.zero_break.problem_specification import DataInfo


class CategoricalStats:
    def __init__(self, data_info: DataInfo, max_print=None):
        self.data_info = data_info
        self.max_print = max_print
        self.cols = self.data_info.feature_types_val["categorical"]
        self.data = self.data_info.data[self.cols]
        self.cols_sorted = self.sort_by_cardinality()

    def any_categorical(self):
        return True if len(self.cols) > 0 else False

    def sort_by_cardinality(self):
        if not self.any_categorical():
            return []
        cardinalities = []
        for col in self.cols:
            data = self.data[col]
            cardinalities.append((col, data.nunique()))
        sorted_cardinalities = sorted(cardinalities, key=lambda x: x[1], reverse=True)
        return [x[0] for x in sorted_cardinalities]

    def sample_top_features(self, n):
        top_features = self.cols_sorted[:3 * n]
        return top_features

    def plot_value_counts(self, n=25, norm=True, plot=True):
        if not self.any_categorical():
            return None
        if not plot:
            return None
        top_n = self.cols_sorted[:min(len(self.cols_sorted), n)]
        num_plots = len(top_n)
        if num_plots == 1:
            col = top_n[0]
            fig, ax = plt.subplots(1, 1, figsize=(20, 5))
            cat_col = self.data[col].apply(lambda x: "cat " + str(x))
            data = cat_col.value_counts(normalize=norm)
            sns.barplot(x=data.index[:5], y=data.values[:5], ax=ax)
            plot_title = f"{col}"
            ax.set_title(plot_title)
            ax.set_xlabel(None)
            ax.set_ylim(0, 1)
            return None
        else:
            dim = math.ceil(math.sqrt(len(top_n)))
            num_rows = math.ceil(num_plots / dim)
            fig, ax = plt.subplots(num_rows, dim)
            fig.tight_layout()
            if norm:
                fig.suptitle("Top 5 Value % for Each Feature")
            else:
                fig.suptitle("Top 5 Value Counts for Each Feature")
            for i, col in enumerate(top_n):
                data = pd.DataFrame(self.data[col].value_counts(normalize=norm)[0:5]).reset_index(drop=False)
                plot_title = f"{col}"
                ax_i = ax[i // dim, i % dim]
                sns.barplot(x=data.iloc[:, 0], y=data.iloc[:, 1].to_list(), ax=ax_i)
                ax_i.set_title(plot_title)
                ax_i.set_xlabel(None)
                if norm:
                    ax_i.set_ylim(0, 1)
            plt.tight_layout()
            plt.show()
            return None

    def calculate_stats(self, is_print=True):
        if not self.any_categorical():
            return None
        result_dicts = []
        for col in self.cols:
            data = self.data[col]
            unique_values = data.nunique()
            missing_values = data.isnull().sum()
            value_counts = data.value_counts().rename("count").reset_index()
            value_counts["percentage"] = 100 * value_counts["count"] / value_counts["count"].sum()
            value_counts = value_counts.sort_values("count", ascending=False)
            dist_str = ', '.join([f"{row['index']}: {row['percentage']:.0f}%" for _, row in value_counts.iterrows()])
            result_dicts.append(
                {'col': col, '# unique': unique_values, '# missing': missing_values, 'Distribution': dist_str})
            data_unique_values = data.dropna().drop_duplicates()
            col_len = len(data_unique_values)
            values_to_sample = 3
            if col_len < values_to_sample:
                values_to_sample = col_len
            sample_values = [str(value) for value in data_unique_values.sample(min(col_len, values_to_sample)).values]
            result_dicts[-1]['Sample'] = ', '.join(sample_values)

            # add more columns
            # 1. % from all that covers the top 5 values
            top_5 = 100 * value_counts.iloc[:5]["count"].sum() / value_counts["count"].sum()
            result_dicts[-1][f'Top 5 values coverage'] = f"{top_5:.0f}%"

        results_df = pd.DataFrame(result_dicts)

        if is_print:
            if self.max_print is not None:
                print(format_df(df=results_df,
                                max_value_width=self.max_print,
                                ))
            else:
                print(format_df(results_df))
        return results_df

    def run(self, plot=True):
        self.calculate_stats()
        self.plot_value_counts(norm=True, plot=plot)
        self.plot_value_counts(norm=False, plot=plot)
        return None


def format_df(df, max_value_width=30):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    df = df.applymap(trim_value)

    return df



from chester.model_training.data_preparation import CVData
from chester.model_training.models.chester_models import best_catboost
from chester.model_training.models.chester_models import best_logistic_regression
from chester.model_training.models.chester_models.base_model import BaseModel
from chester.model_training.models.chester_models.best_baseline_model import BaselineModel
from chester.model_training.models.chester_models.best_linear_regression import LinearRegressionModel
from chester.model_training.models.chester_models.catboost.catboost_utils import compare_best_models
from chester.zero_break.problem_specification import DataInfo


class BestModel(BaseModel):
    def __init__(self, data_info: DataInfo, cv_data: CVData, num_models_to_compare=15, best_practice_prob=0.33):
        super().__init__(data_info, cv_data, num_models_to_compare, best_practice_prob)

    def get_best_model(self):
        models_type = self.data_info.model_selection_val
        models = []
        for model_type in models_type:
            if 'baseline' in model_type:
                base_res, model = BaselineModel(data_info=self.data_info,
                                                cv_data=self.cv_data,
                                                num_models_to_compare=3
                                                ).get_best_model()
                models.append((base_res, model))
            elif 'logistic' in model_type:
                base_res, model = best_logistic_regression.LogisticRegressionModel(
                    data_info=self.data_info,
                    cv_data=self.cv_data,
                    num_models_to_compare=self.num_models_to_compare,
                    best_practice_prop=self.best_practice_prop
                ).get_best_model()
                models.append((base_res, model))
            elif 'catboost' in model_type:
                base_res, model = best_catboost.CatboostModel(
                    data_info=self.data_info,
                    cv_data=self.cv_data,
                    num_models_to_compare=self.num_models_to_compare
                ).get_best_model()
                models.append((base_res, model))
            elif 'linear' in model_type:
                base_res, model = LinearRegressionModel(
                    data_info=self.data_info,
                    cv_data=self.cv_data,
                    num_models_to_compare=self.num_models_to_compare,
                    best_practice_prob=self.best_practice_prop
                ).get_best_model()
                models.append((base_res, model))
        if models is None:
            return None
        else:
            print("Finding the best model out of model types ran")
            best = compare_best_models(models)
            return best



import matplotlib.pyplot as plt
import numpy as np
from pandas.errors import SettingWithCopyWarning

from chester.feature_stats.categorical_stats import CategoricalStats
from chester.feature_stats.numeric_stats import NumericStats
from chester.zero_break.problem_specification import DataInfo


class TargetPreModelAnalysis:
    def __init__(self, data_info: DataInfo):
        import warnings
        warnings.filterwarnings("ignore", category=SettingWithCopyWarning)
        self.data_info = data_info
        self.target = self.data_info.data[self.data_info.target]
        self.target_col = self.data_info.target
        self.target_df = self.data_info.data[[self.data_info.target]]
        self.target_df.rename(columns={self.data_info.target: 'target_label'}, inplace=True)
        self.data_info = DataInfo(data=self.target_df)
        self.data_info.calculate()
        self.target_type = None
        # target type
        for feature_type, features in self.data_info.feature_types_val.items():
            if 'target_label' in features:
                self.target_type = feature_type

    def plot_histogram(self):
        target = self.target
        plt.hist(target, bins=30, edgecolor='k')
        plt.figure(figsize=(10, 10))
        # plt.rcParams.update({'font.size': 12})
        plt.xlabel('Values')
        plt.ylabel('Counts')
        plt.title(f'Histogram of {self.target.name}')
        plt.show()
        plt.close()

    def plot_barplot(self):
        target = self.target
        value_counts = target.value_counts()
        percentages = np.array(value_counts / target.size * 100)
        fig, ax1 = plt.subplots()
        # plt.figure(figsize=(10, 10))
        # plt.rcParams.update({'font.size': 12})
        ax2 = ax1.twinx()
        ax1.bar(value_counts.index, value_counts.values, color='gray')
        ax1.set_ylabel('Counts', color='gray')
        ax2.plot(value_counts.index, percentages, color='red', marker='o')
        ax2.set_ylabel('Percentages', color='red')
        ax1.set_xlabel('Values')
        plt.title(f'Bar Plot of {self.target.name}')
        plt.show()

    def run(self, plot=True):
        if self.target_type == "numeric":
            print("Numerical statistics for target column")
            NumericStats(self.data_info).run(plot=False)
            if plot:
                self.plot_histogram()
        elif self.target_type == "categorical":
            print("Categorical statistics for target column")
            CategoricalStats(self.data_info).run(plot=False)
            if plot:
                self.plot_barplot()



from sklearn.compose import ColumnTransformer
from sklearn.exceptions import ConvergenceWarning
from sklearn.impute import SimpleImputer
from sklearn.linear_model import ElasticNet
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder

from chester.zero_break.problem_specification import DataInfo


class LinearRegressionModel:
    def __init__(self, parameters: list, data_info: DataInfo):
        self.parameters = parameters
        self.data_info = data_info
        self.model_type = self.data_info.problem_type_val
        if "regression" in self.model_type.lower():
            self.en_model = ElasticNet()
        self.categorical_features = self.data_info.feature_types_val["categorical"]
        self.numeric_features = self.data_info.feature_types_val["numeric"]
        self.transformer = None
        self.pipeline = None

    def fit(self, X, y):
        import warnings
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        # Prepare the numerical features for processing
        numerical_transformer = SimpleImputer(strategy='median')
        # Prepare the categorical features for processing
        categorical_transformer = OneHotEncoder(handle_unknown='ignore')
        # Use the ColumnTransformer to apply the transformations to the corresponding features
        self.transformer = ColumnTransformer(
            transformers=[
                ('c', categorical_transformer, self.categorical_features),
                ('n', numerical_transformer, self.numeric_features)
            ])
        # Prepare the pipeline
        self.pipeline = Pipeline(steps=[('preprocessor', self.transformer),
                                        ('regressor', self.en_model)
                                        ])
        # Fit the pipeline with the training data
        hyperparams = {param.name: param.value for param in self.parameters}
        self.en_model.set_params(**hyperparams)
        self.pipeline.fit(X, y)

    def retrain(self, X, y):
        # Fit the pipeline with the updated data
        self.pipeline.fit(X, y)

    def predict(self, X):
        # Use the pipeline to predict the target values
        return self.pipeline.predict(X)

    def transform(self, X):
        return self.predict(X)

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)

    def get_params(self):
        return self.parameters



from chester.text_stats_analysis.data_quality import TextAnalyzer
from chester.text_stats_analysis.smart_text_analyzer import analyze_text_df
from chester.zero_break.problem_specification import DataInfo


class TextStats:
    def __init__(self, data_info: DataInfo):
        self.data_info = data_info
        self.cols = self.data_info.feature_types_val["text"]
        self.data = self.data_info.data  # [self.cols]

    def run(self):
        for col in self.cols:
            print(f"** Analayzing {col} column")
            text_analyzer = TextAnalyzer(df=self.data, text_column=col)
            analyze_text_df(text_analyzer)



from chester.zero_break.problem_specification import DataInfo


class FeatureStatistics:
    def __init__(self, data_info: DataInfo):
        self.data_info = data_info

    def calculate_stats(self):
        pass



import pandas as pd


class BaselineModel:
    def __init__(self,
                 baseline_value=None,
                 mode_baseline=False,
                 median_baseline=False,
                 avg_baseline=False):
        self.current_baseline = baseline_value
        self.mode_baseline = mode_baseline
        self.median_baseline = median_baseline
        self.avg_baseline = avg_baseline

    def fit(self, y):
        if self.current_baseline is None:
            if self.avg_baseline:
                self.current_baseline = y.mean()
            elif self.median_baseline:
                self.current_baseline = y.quantile(0.5)
            else:
                self.current_baseline = y.mode()[0]

    def transform(self, X):
        return pd.Series([self.current_baseline] * len(X))

    def predict(self, X):
        return self.transform(X)

    def fit_transform(self, X, y):
        self.fit(y)
        return self.transform(X)



import pandas as pd

from chester.features_engineering.feature_handler import FeatureHandler
from chester.run.user_classes import TextFeatureSpec
from chester.zero_break.problem_specification import DataInfo


class FeaturesHandler:
    def __init__(self, data_info: DataInfo, text_feature_extraction: TextFeatureSpec = None):
        self.data = data_info.data
        self.target = data_info.target
        self.text_feature_extraction = text_feature_extraction
        self.problem_type_val = data_info.problem_type_val
        self.feature_types_val = data_info.feature_types_val
        self.loss_detector_val = data_info.loss_detector_val
        self.metrics_detector_val = data_info.metrics_detector_val
        self.model_selection_val = data_info.model_selection_val
        self.label_transformation_val = data_info.label_transformation_val

    def _get_features_handler(self, data: pd.DataFrame = None):
        if data is not None:
            data = self.data
        feature_handlers = []
        for col in data.columns:
            feature_type = None
            for key, value in self.feature_types_val.items():
                if col in value:
                    feature_type = key
                    break
                # Prepare an instance of FeatureHandler
            feature_handler = FeatureHandler(
                column=data[col],
                feature_type=feature_type,
                col_name=col,
                text_feature_extraction=self.text_feature_extraction)
            feature_handlers.append(feature_handler)
        return feature_handlers

    def transform(self):
        feature_types = {'numeric': [], 'categorical': []}
        feature_handlers = self._get_features_handler(data=self.data)
        print(f"Handling {len(feature_handlers)} raw features")
        feat_values = []
        feat_names = []
        for feature_handler in feature_handlers:
            try:
                values, names = feature_handler.handle_feature()
                feat_values.append(values)
                if feature_handler.feature_type is None:
                    pass
                if feature_handler.feature_type == 'numeric':
                    feature_types['numeric'].extend(names)
                    feat_names.append(names)
                elif feature_handler.feature_type == 'boolean':
                    feature_types['numeric'].extend(names)
                    feat_names.append(names)
                elif feature_handler.feature_type == 'text':
                    feature_types['numeric'].extend(names)
                    feat_names.append(names)
                elif feature_handler.feature_type == 'categorical':
                    feat_names.append(names)
                    feature_types['categorical'].extend(names)
            except:
                pass
        final_df = pd.DataFrame()
        for value in feat_values:
            if type(value) == pd.DataFrame:
                final_df = pd.concat([final_df, value], axis=1)
            elif type(value) == pd.Series:
                value = value.to_frame().reset_index(drop=True)
                final_df = pd.concat([final_df, value], axis=1)
        return feature_types, final_df



from chester.model_training.data_preparation import CVData
from chester.model_training.models.chester_models.base_model_utils import calculate_metrics_scores
from chester.model_training.models.chester_models.base_model_utils import is_metric_higher_is_better
from chester.model_training.models.chester_models.baseline.baseline_model import BaselineModel


def train_baseline(X_train, y_train,
                   baseline_value=None,
                   mode_baseline=False,
                   median_baseline=False,
                   avg_baseline=False):
    """
    Trains a baseline model using the given parameters.
    :param X_train: The training data features (unused in this function)
    :param y_train: The training data labels
    :param baseline_num: A number to use as the baseline (mean, median, mode, or percentile)
    :param percentile: Percentile to use as the baseline if baseline_num is not provided
    :return: A trained baseline model
    """
    model = BaselineModel(baseline_value=baseline_value, mode_baseline=mode_baseline,
                          median_baseline=median_baseline, avg_baseline=avg_baseline)
    model.fit(y_train)
    return model


def predict_baseline(model, X):
    """
    Makes predictions using a baseline model
    :param model: trained baseline model
    :param X: The data to make predictions on (unused in this function)
    :return: A list of predictions
    """
    return model.transform(X)


def baseline_with_outputs(cv_data: CVData,
                          metrics: list,
                          target_col: str,
                          baseline_value=None,
                          mode_baseline=False,
                          median_baseline=False,
                          avg_baseline=False,
                          ):
    results = []
    for i, (train_index, test_index) in enumerate(cv_data.splits):
        X_train, X_test = cv_data.train_data.iloc[train_index], cv_data.train_data.iloc[test_index]
        y_train, y_test = X_train[target_col], X_test[target_col]
        model = train_baseline(X_train, y_train, baseline_value=baseline_value, mode_baseline=mode_baseline,
                               median_baseline=median_baseline, avg_baseline=avg_baseline)
        prediction = predict_baseline(model, X_test)
        prediction_train = predict_baseline(model, X_train)
        scores = calculate_metrics_scores(y_test, prediction, metrics)
        results.append({'type': 'test', 'fold': i, **scores})
        scores = calculate_metrics_scores(y_train, prediction_train, metrics)
        results.append({'type': 'train', 'fold': i, **scores})
    return results, model


import pandas as pd


def compare_models(results):
    all_results = [(pd.DataFrame(result), model) for result, model in results]
    metric_name = [col for col in all_results[0][0].columns if col not in ['type', 'fold']][0]
    sort_ascending = is_metric_higher_is_better(metric_name)
    best_result = None
    best_model = None
    best_value = None
    for (result, model) in all_results:
        test_result = result[result['type'] == 'test'].groupby('fold').mean(numeric_only=True).reset_index()
        mean_value = test_result[metric_name].mean()
        if best_value is None or \
                (sort_ascending and mean_value > best_value) \
                or (not sort_ascending and mean_value < best_value):
            best_value = mean_value
            best_result = result
            best_model = model
    return best_result, best_model



import random
from typing import List

import numpy as np
import pandas as pd

from chester.model_training.data_preparation import CVData
from chester.model_training.data_preparation import Parameter
from chester.model_training.models.chester_models.base_model_utils import calculate_metrics_scores
from chester.model_training.models.chester_models.base_model_utils import is_metric_higher_is_better
from chester.model_training.models.chester_models.hp_generator import HPGenerator
from chester.model_training.models.chester_models.logistic_regression.logistic_regression_model import \
    LogisticRegressionModel
from chester.zero_break.problem_specification import DataInfo


def train_logistic_regression(X_train, y_train, parameters: list, data_info: DataInfo):
    """
    Trains a baseline model using the given parameters.
    :param X_train: The training data features (unused in this function)
    :param y_train: The training data labels
    :param parameters: parameters list
    :return: A trained baseline model
    """
    model = LogisticRegressionModel(parameters, data_info)
    model.fit(X_train, y_train)
    return model


def predict_logistic_regression(model, X):
    """
    Makes predictions using a baseline model
    :param model: trained baseline model
    :param X: The data to make predictions on (unused in this function)
    :return: A list of predictions
    """
    return model.transform(X)


def logistic_regression_with_outputs(cv_data: CVData,
                                     metrics: list,
                                     target_col: str,
                                     parameters: list,
                                     data_info: DataInfo,
                                     problem_type=None,
                                     ):
    results = []
    for i, (X_train, y_train, X_test, y_test) in enumerate(cv_data.format_splits()):
        model = train_logistic_regression(X_train, y_train, parameters=parameters, data_info=data_info)
        prediction = predict_logistic_regression(model, X_test)
        prediction_train = predict_logistic_regression(model, X_train)
        scores = calculate_metrics_scores(y_test, prediction, metrics, problem_type)
        results.append({'type': 'test', 'fold': i, **scores})
        scores = calculate_metrics_scores(y_train, prediction_train, metrics, problem_type)
        results.append({'type': 'train', 'fold': i, **scores})
    return results, model


def compare_models(results):
    all_results = [(pd.DataFrame(result), model) for result, model in results]
    # print("all_results", all_results[0][0])
    metric_name = [col for col in all_results[0][0].columns if col not in ['type', 'fold']][0]
    sort_ascending = is_metric_higher_is_better(metric_name)
    best_result = None
    best_model = None
    best_value = None
    for (result, model) in all_results:
        test_result = result[result['type'] == 'test'].groupby('fold').mean(numeric_only=True).reset_index()
        mean_value = test_result[metric_name].mean()
        if best_value is None or \
                (sort_ascending and mean_value > best_value) \
                or (not sort_ascending and mean_value < best_value):
            best_value = mean_value
            best_result = result
            best_model = model
    return best_result, best_model


default_parameters = {
    'penalty': 'l2',
    'C': 1.0,
    'solver': 'lbfgs',
    'max_iter': 100
}

logistic_regression_default_parameters = [
    Parameter('penalty', default_parameters['penalty']),
    Parameter('C', default_parameters['C']),
    Parameter('solver', default_parameters['solver']),
    Parameter('max_iter', default_parameters['max_iter'])
]


def generate_logistic_regression_configs(k: int = 10, best_practice_prob=0.33) -> List[List[Parameter]]:
    hp_generator = HPGeneratorLR(n_models=k, best_practice_prob=best_practice_prob)
    parameter_format = hp_generator.hp_format(hp_generator.generate_configs())
    return parameter_format


class HPGeneratorLR(HPGenerator):
    def __init__(self, best_practice_configs: list = None,
                 categorical_sample_configs: list = None,
                 n_models=9,
                 best_practice_prob=0.33):
        super().__init__(best_practice_configs, categorical_sample_configs, n_models, best_practice_prob)
        self.best_practice_configs = self.load_best_practice_configs()

    @staticmethod
    def load_best_practice_configs():
        return [
            {'penalty': 'l1', 'C': 0.1, 'solver': 'saga'},
            {'penalty': 'l2', 'C': 0.1, 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'C': 0.1, 'solver': 'saga', 'l1_ratio': 0.5},
            {'penalty': 'l2', 'C': 0.01, 'solver': 'lbfgs'},
            {'penalty': 'l1', 'C': 0.5, 'solver': 'saga'},
            {'penalty': 'l2', 'C': 0.5, 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'C': 0.5, 'solver': 'saga', 'l1_ratio': 0.5},
            {'penalty': 'l2', 'C': 0.001, 'solver': 'lbfgs'},
            {'penalty': 'l1', 'C': 1, 'solver': 'saga'},
            {'penalty': 'l2', 'C': 1, 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'C': 1, 'solver': 'saga', 'l1_ratio': 0.5},
        ]

    def generate_random_config(self) -> dict:
        cat_configs = [
            {'penalty': 'l1', 'solver': 'saga'},
            {'penalty': 'l2', 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'solver': 'saga', 'l1_ratio': 0.5},
            {'penalty': 'l2', 'solver': 'lbfgs'},
            {'penalty': 'l1', 'solver': 'saga'},
            {'penalty': 'l2', 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'solver': 'saga', 'l1_ratio': 0.5},
            {'penalty': 'l2', 'solver': 'lbfgs'},
            {'penalty': 'l1', 'solver': 'saga'},
            {'penalty': 'l2', 'solver': 'newton-cg'},
            {'penalty': 'elasticnet', 'solver': 'saga', 'l1_ratio': 0.5},
        ]

        cat_config = random.choice(cat_configs)
        config = cat_config.copy()
        config['C'] = self.generate_c_value()
        if 'l1_ratio' in cat_config:
            config['l1_ratio'] = self.generate_l1_ratio_random_configs()

        return config

    @staticmethod
    def generate_l1_ratio_random_configs() -> float:
        l1_ratio_values = [0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99]
        return random.choice(l1_ratio_values)

    @staticmethod
    def generate_c_value():
        choice = random.choice([0, 1])
        if choice == 0:
            return np.random.uniform(0.0001, 100)
        else:
            return np.random.uniform(np.log10(1.001), np.log10(100))



from __future__ import division

import operator
import string

import nltk


def isPunct(word):
    return len(word) == 1 and word in string.punctuation


def isNumeric(word):
    try:
        float(word) if '.' in word else int(word)
        return True
    except ValueError:
        return False


class RakeKeywordExtractor:

    def __init__(self):
        self.stopwords = set(nltk.corpus.stopwords.words())
        self.top_fraction = 1  # consider top third candidate keywords by score

    def _generate_candidate_keywords(self, sentences):
        phrase_list = []
        for sentence in sentences:
            words = map(lambda x: "|" if x in self.stopwords else x,
                        nltk.word_tokenize(sentence.lower()))
            phrase = []
            for word in words:
                if word == "|" or isPunct(word):
                    if len(phrase) > 0:
                        phrase_list.append(phrase)
                        phrase = []
                else:
                    phrase.append(word)
        return phrase_list

    def _calculate_word_scores(self, phrase_list):
        word_freq = nltk.FreqDist()
        word_degree = nltk.FreqDist()
        for phrase in phrase_list:
            degree = len(list(filter(lambda x: not isNumeric(x), phrase))) - 1
            for word in phrase:
                word_freq.update([word])
                word_degree.update([word])
                # word_degree.inc(word, degree)  # other words
        for word in word_freq.keys():
            word_degree[word] = word_degree[word] + word_freq[word]  # itself
        # word score = deg(w) / freq(w)
        word_scores = {}
        for word in word_freq.keys():
            word_scores[word] = word_degree[word] / word_freq[word]
        return word_scores

    def _calculate_phrase_scores(self, phrase_list, word_scores):
        phrase_scores = {}
        for phrase in phrase_list:
            phrase_score = 0
            for word in phrase:
                phrase_score += word_scores[word]
            phrase_scores[" ".join(phrase)] = phrase_score
        return phrase_scores

    def extract(self, text, max_words=7, top_words=10):
        sentences = nltk.sent_tokenize(text)
        phrase_list = self._generate_candidate_keywords(sentences)
        word_scores = self._calculate_word_scores(phrase_list)
        phrase_scores = self._calculate_phrase_scores(
            phrase_list, word_scores)
        sorted_phrase_scores = sorted(phrase_scores.items(), key=operator.itemgetter(1), reverse=True)
        n_phrases = len(sorted_phrase_scores)
        return truncate_and_limit_keywords(sorted_phrase_scores, max_words, top_words)


def truncate_and_limit_keywords(keywords_list, max_words, top_words):
    truncated_keywords = []
    for keyword in keywords_list[:top_words]:
        text = trim_repeated_words(keyword[0], 1)
        score = keyword[1]
        truncated_text = " ".join(text.split()[:max_words])
        truncated_keywords.append((truncated_text, score))
    return truncated_keywords


def trim_repeated_words(text, max_repeated_words):
    words = text.split()
    new_words = []
    word_count = {}
    for word in words:
        if word in word_count:
            word_count[word] += 1
            if word_count[word] <= max_repeated_words:
                new_words.append(word)
        else:
            word_count[word] = 1
            new_words.append(word)
    return ' '.join(new_words)



import random

from chester.model_training.data_preparation import Parameter


class HPGenerator:
    def __init__(self,
                 best_practice_configs: list = None,
                 categorical_sample_configs: list = None,
                 n_models=9,
                 best_practice_prob=0.33):
        self.n_models = n_models
        self.best_practice_prob = best_practice_prob
        self.best_practice_configs = best_practice_configs
        self.categorical_sample_configs = categorical_sample_configs

    def generate_configs(self):
        configs = []
        for i in range(self.n_models):
            configs.append(self.generate_config())
        return configs

    def generate_config(self):
        if random.uniform(0, 1) < self.best_practice_prob:
            return self.generate_bp_config()
        else:
            return self.generate_random_config()

    def generate_bp_config(self):
        return random.choice(self.best_practice_configs)

    def generate_categorical_random_config(self):
        return random.choice(self.categorical_sample_configs)

    def generate_numerics_random_config(self) -> dict:
        pass

    def generate_random_config(self):
        categorical_config = self.generate_categorical_random_config()
        numerics_config = self.generate_numerics_random_config()
        combined_config = {**categorical_config, **numerics_config}
        return combined_config

    @staticmethod
    def hp_format(hp_list):

        hp_list_formatted = []

        for hp in hp_list:
            parameters = []
            for name, value in hp.items():
                parameters.append(Parameter(name, value))
            hp_list_formatted.append(parameters)
        return hp_list_formatted



from chester.model_training.data_preparation import CVData
from chester.model_training.models.chester_models.base_model import BaseModel
from chester.model_training.models.chester_models.logistic_regression.logistic_regression_utils import \
    generate_logistic_regression_configs, logistic_regression_with_outputs, compare_models
from chester.zero_break.problem_specification import DataInfo


class LogisticRegressionModel(BaseModel):
    def __init__(self, data_info: DataInfo, cv_data: CVData, num_models_to_compare=9, best_practice_prop=0.33):
        super().__init__(data_info, cv_data, num_models_to_compare)
        self.hp_list = generate_logistic_regression_configs(k=self.num_models_to_compare,
                                                            best_practice_prob=best_practice_prop)
        print(f"Running {self.num_models_to_compare} elastic net models")

    def get_best_model(self):
        models = self.data_info.model_selection_val
        metrics = self.get_metrics_functions()
        if models is None:
            return None
        else:
            models = [model for model in models if "logistic" in model]
            if len(models) == 0:
                return None
            else:
                results = []
                for _ in models:
                    for params in self.hp_list:
                        base_res, model = logistic_regression_with_outputs(
                            cv_data=self.cv_data, target_col=self.cv_data.target_column,
                            parameters=params, metrics=metrics, data_info=self.data_info,
                            problem_type=self.data_info.problem_type_val)
                        results.append((base_res, model))
                best = compare_models(results)
                return best

    def calc_model_score(self, model):
        # calculate the score of a given baseline model
        pass



from chester.model_training.data_preparation import CVData
from chester.model_training.models.chester_models.base_model import BaseModel
from chester.model_training.models.chester_models.baseline.baseline_utils import baseline_with_outputs, compare_models
from chester.zero_break.problem_specification import DataInfo


class BaselineModel(BaseModel):
    def __init__(self, data_info: DataInfo, cv_data: CVData, num_models_to_compare=10):
        super().__init__(data_info, cv_data, num_models_to_compare)

    def get_best_model(self):
        models = self.data_info.model_selection_val
        metrics = self.get_metrics_functions()
        if models is None:
            return None
        else:
            baseline_models = [model.split("-")[1] for model in models if "baseline" in model]
            if len(baseline_models) == 0:
                return None
            else:
                results = []
                for baseline_model in baseline_models:
                    median_baseline = None
                    if baseline_model == 'median':
                        median_baseline = 0.5

                    average_baseline = False
                    if baseline_model == 'average':
                        average_baseline = True

                    baseline_value = None
                    if 'value' in baseline_model:
                        baseline_value = baseline_model.split(" ")[1]
                    base_res, model = baseline_with_outputs(
                        cv_data=self.cv_data, target_col=self.cv_data.target_column,
                        baseline_value=baseline_value, avg_baseline=average_baseline, median_baseline=median_baseline,
                        metrics=metrics)
                    results.append((base_res, model))
                best = compare_models(results)
                return best

    def calc_model_score(self, model):
        # calculate the score of a given baseline model
        pass
